package com.alibaba.spring.boot.rsocket.broker.upstream;

import com.alibaba.rsocket.RSocketAppContext;
import com.alibaba.rsocket.RSocketRequesterSupport;
import com.alibaba.rsocket.ServiceLocator;
import com.alibaba.rsocket.events.ServicesExposedEvent;
import com.alibaba.rsocket.metadata.AppMetadata;
import com.alibaba.rsocket.metadata.BearerTokenMetadata;
import com.alibaba.rsocket.metadata.RSocketCompositeMetadata;
import com.alibaba.rsocket.route.RSocketFilterChain;
import com.alibaba.rsocket.transport.NetworkUtil;
import com.alibaba.spring.boot.rsocket.broker.RSocketBrokerProperties;
import com.alibaba.spring.boot.rsocket.broker.responder.RSocketBrokerHandlerRegistry;
import com.alibaba.spring.boot.rsocket.broker.route.ServiceRoutingSelector;
import io.cloudevents.v1.CloudEventBuilder;
import io.cloudevents.v1.CloudEventImpl;
import io.netty.buffer.Unpooled;
import io.rsocket.Payload;
import io.rsocket.SocketAcceptor;
import io.rsocket.metadata.WellKnownMimeType;
import io.rsocket.plugins.RSocketInterceptor;
import io.rsocket.util.ByteBufPayload;
import org.springframework.core.env.Environment;
import reactor.core.publisher.Mono;

import java.net.URI;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.function.Supplier;
import java.util.stream.Collectors;

/**
 * RSocket Requester by sub Broker
 *
 * @author leijuan
 */
public class RSocketRequesterBySubBroker implements RSocketRequesterSupport {
    private RSocketBrokerProperties properties;
    private Environment env;
    private String appName;
    private char[] jwtToken;
    private ServiceRoutingSelector serviceRoutingSelector;
    private RSocketBrokerHandlerRegistry handlerRegistry;
    private RSocketFilterChain filterChain;
    private UpstreamBrokerCluster upstreamBrokerCluster;

    public RSocketRequesterBySubBroker(RSocketBrokerProperties properties, Environment env,
                                       RSocketBrokerHandlerRegistry handlerRegistry,
                                       RSocketFilterChain filterChain,
                                       ServiceRoutingSelector serviceRoutingSelector) {
        this.env = env;
        this.appName = env.getProperty("spring.application.name", env.getProperty("application.name", "unknown-app"));
        this.jwtToken = env.getProperty("rsocket.jwt-token", "").toCharArray();
        this.serviceRoutingSelector = serviceRoutingSelector;
        this.handlerRegistry = handlerRegistry;
        this.filterChain = filterChain;
        this.properties = properties;
    }

    public void setUpstreamBrokerCluster(UpstreamBrokerCluster upstreamBrokerCluster) {
        this.upstreamBrokerCluster = upstreamBrokerCluster;
    }

    @Override
    public URI originUri() {
        return URI.create("tcp://" + NetworkUtil.LOCAL_IP + ":" + properties.getPort()
                + "?appName=" + appName
                + "&uuid=" + RSocketAppContext.ID);
    }

    @Override
    public Supplier<Payload> setupPayload() {
        return () -> {
            //composite metadata with app metadata
            RSocketCompositeMetadata compositeMetadata = RSocketCompositeMetadata.from(getAppMetadata());
            // authentication
            if (this.jwtToken != null && this.jwtToken.length > 0) {
                compositeMetadata.addMetadata(new BearerTokenMetadata(this.jwtToken));
            }
            return ByteBufPayload.create(Unpooled.EMPTY_BUFFER, compositeMetadata.getContent());
        };
    }

    @Override
    public Supplier<Set<ServiceLocator>> exposedServices() {
        return () -> {
            return serviceRoutingSelector.findAllServices().stream()
                    .filter(serviceLocator -> serviceLocator.hasTag("global"))
                    .collect(Collectors.toSet());
        };
    }

    @Override
    public Supplier<Set<ServiceLocator>> subscribedServices() {
        return Collections::emptySet;
    }

    @Override
    public Supplier<CloudEventImpl<ServicesExposedEvent>> servicesExposedEvent() {
        return () -> {
            Collection<ServiceLocator> serviceLocators = exposedServices().get();
            if (serviceLocators.isEmpty()) return null;
            ServicesExposedEvent servicesExposedEvent = new ServicesExposedEvent();
            for (ServiceLocator serviceLocator : serviceLocators) {
                servicesExposedEvent.addService(serviceLocator);
            }
            servicesExposedEvent.setAppId(RSocketAppContext.ID);
            return CloudEventBuilder.<ServicesExposedEvent>builder()
                    .withId(UUID.randomUUID().toString())
                    .withTime(ZonedDateTime.now())
                    .withSource(URI.create("broker://" + RSocketAppContext.ID))
                    .withType(ServicesExposedEvent.class.getCanonicalName())
                    .withDataContentType(WellKnownMimeType.APPLICATION_JSON.getString())
                    .withData(servicesExposedEvent)
                    .build();
        };
    }

    @Override
    public SocketAcceptor socketAcceptor() {
        return (connectionSetupPayload, rsocket) -> Mono.just(new UpstreamForwardRSocket(this.serviceRoutingSelector, rsocket, this.filterChain, this.handlerRegistry, this.upstreamBrokerCluster));
    }

    @Override
    public List<RSocketInterceptor> responderInterceptors() {
        return Collections.emptyList();
    }

    @Override
    public List<RSocketInterceptor> requestInterceptors() {
        return Collections.emptyList();
    }

    private AppMetadata getAppMetadata() {
        //app metadata
        AppMetadata appMetadata = new AppMetadata();
        appMetadata.setUuid(RSocketAppContext.ID);
        appMetadata.setName(appName);
        appMetadata.setIp(NetworkUtil.LOCAL_IP);
        appMetadata.setDevice("SpringBootApp");
        //rsocket schema
        appMetadata.setSchema("tcp");
        //rsocket port
        appMetadata.setPort(properties.getPort());
        //brokers
        appMetadata.setBrokers(properties.getUpstreamBrokers());
        appMetadata.setTopology(properties.getTopology());
        //management port
        appMetadata.setManagementPort(Integer.parseInt(env.getProperty("management.server.port",
                env.getProperty("server.port", "8080"))));
        appMetadata.addMetadata("broker", "true");
        return appMetadata;
    }
}
